import re
import sys
import uuid
import copy
import datetime

from six import iteritems

from . import RefinerUtils
from pandajedi.jedicore import Interaction
from pandajedi.jedicore import JediException
from pandajedi.jedicore.JediTaskSpec import JediTaskSpec
from pandajedi.jedicore.JediDatasetSpec import JediDatasetSpec
from pandajedi.jedicore.JediFileSpec import JediFileSpec
from pandaserver.taskbuffer import EventServiceUtils

try:
    from idds.client.client import Client as iDDS_Client
    import idds.common.constants
    import idds.common.utils
except ImportError:
    pass


# base class for task refine
class TaskRefinerBase (object):

    # constructor
    def __init__(self,taskBufferIF,ddmIF):
        self.ddmIF = ddmIF
        self.taskBufferIF = taskBufferIF
        self.initializeRefiner(None)
        self.refresh()



    # refresh
    def refresh(self):
        self.siteMapper = self.taskBufferIF.getSiteMapper()



    # initialize
    def initializeRefiner(self,tmpLog):
        self.taskSpec = None
        self.inMasterDatasetSpec = []
        self.inSecDatasetSpecList = []
        self.outDatasetSpecList = []
        self.outputTemplateMap = {}
        self.jobParamsTemplate = None
        self.cloudName = None
        self.siteName = None
        self.tmpLog = tmpLog
        self.updatedTaskParams = None
        self.unmergeMasterDatasetSpec = {}
        self.unmergeDatasetSpecMap = {}
        self.oldTaskStatus = None
        self.unknownDatasetList = []



    # set jobParamsTemplate
    def setJobParamsTemplate(self,jobParamsTemplate):
        self.jobParamsTemplate = jobParamsTemplate



    # extract common parameters
    def extractCommon(self,jediTaskID,taskParamMap,workQueueMapper,splitRule):
        # make task spec
        taskSpec = JediTaskSpec()
        taskSpec.jediTaskID = jediTaskID
        taskSpec.attemptNr = 0
        taskSpec.taskName = taskParamMap['taskName']
        taskSpec.userName = taskParamMap['userName']
        taskSpec.vo = taskParamMap['vo']
        taskSpec.prodSourceLabel = taskParamMap['prodSourceLabel']
        taskSpec.taskPriority = taskParamMap['taskPriority']
        if taskSpec.taskPriority is None:
            taskSpec.taskPriority = 0
        if 'currentPriority' in taskParamMap:
            taskSpec.currentPriority = taskParamMap['currentPriority']
        else:
            taskSpec.currentPriority = taskSpec.taskPriority
        taskSpec.architecture = taskParamMap['architecture']
        taskSpec.transUses = taskParamMap['transUses']
        taskSpec.transHome = taskParamMap['transHome']
        if 'transPath' in taskParamMap:
            taskSpec.transPath = taskParamMap['transPath']
        taskSpec.processingType = taskParamMap['processingType']
        taskSpec.taskType = taskParamMap['taskType']
        taskSpec.splitRule = splitRule
        taskSpec.startTime = datetime.datetime.utcnow()
        if 'workingGroup' in taskParamMap:
            taskSpec.workingGroup = taskParamMap['workingGroup']
        if 'countryGroup' in taskParamMap:
            taskSpec.countryGroup = taskParamMap['countryGroup']
        if 'ticketID' in taskParamMap:
            taskSpec.ticketID = taskParamMap['ticketID']
        if 'ticketSystemType' in taskParamMap:
            taskSpec.ticketSystemType = taskParamMap['ticketSystemType']
        if 'reqID' in taskParamMap:
            taskSpec.reqID = taskParamMap['reqID']
        else:
            taskSpec.reqID = jediTaskID
        if 'coreCount' in taskParamMap:
            taskSpec.coreCount = taskParamMap['coreCount']
        else:
            taskSpec.coreCount = 1
        if 'walltime' in taskParamMap:
            taskSpec.walltime = taskParamMap['walltime']
        else:
            taskSpec.walltime = 0
        if 'walltimeUnit' not in taskParamMap:
            # force to set NULL so that retried tasks get data from scouts again
            taskSpec.forceUpdate('walltimeUnit')
        if 'outDiskCount' in taskParamMap:
            taskSpec.outDiskCount = taskParamMap['outDiskCount']
        else:
            taskSpec.outDiskCount = 0
        if 'outDiskUnit' in taskParamMap:
            taskSpec.outDiskUnit = taskParamMap['outDiskUnit']
        if 'workDiskCount' in taskParamMap:
            taskSpec.workDiskCount = taskParamMap['workDiskCount']
        else:
            taskSpec.workDiskCount = 0
        if 'workDiskUnit' in taskParamMap:
            taskSpec.workDiskUnit = taskParamMap['workDiskUnit']
        if 'ramCount' in taskParamMap:
            taskSpec.ramCount = taskParamMap['ramCount']
        else:
            taskSpec.ramCount = 0
        if 'ramUnit' in taskParamMap:
            taskSpec.ramUnit = taskParamMap['ramUnit']
        if 'baseRamCount' in taskParamMap:
            taskSpec.baseRamCount = taskParamMap['baseRamCount']
        else:
            taskSpec.baseRamCount = 0
        # IO
        if 'ioIntensity' in taskParamMap:
            taskSpec.ioIntensity = taskParamMap['ioIntensity']
        if 'ioIntensityUnit' in taskParamMap:
            taskSpec.ioIntensityUnit = taskParamMap['ioIntensityUnit']
        # HS06 stuff
        if 'cpuTimeUnit' in taskParamMap:
            taskSpec.cpuTimeUnit = taskParamMap['cpuTimeUnit']
        if 'cpuTime' in taskParamMap:
            taskSpec.cpuTime = taskParamMap['cpuTime']
        if 'cpuEfficiency' in taskParamMap:
            taskSpec.cpuEfficiency = taskParamMap['cpuEfficiency']
        else:
            # 90% of cpu efficiency by default
            taskSpec.cpuEfficiency = 90
        if 'baseWalltime' in taskParamMap:
            taskSpec.baseWalltime = taskParamMap['baseWalltime']
        else:
            # 10min of offset by default
            taskSpec.baseWalltime = 10*60
        # for merge
        if 'mergeRamCount' in taskParamMap:
            taskSpec.mergeRamCount = taskParamMap['mergeRamCount']
        if 'mergeCoreCount' in taskParamMap:
            taskSpec.mergeCoreCount = taskParamMap['mergeCoreCount']
        # scout
        if 'skipScout' not in taskParamMap and not taskSpec.isPostScout():
            taskSpec.setUseScout(True)
        # cloud
        if 'cloud' in taskParamMap:
            self.cloudName = taskParamMap['cloud']
            taskSpec.cloud = self.cloudName
        else:
            # set dummy to force update
            taskSpec.cloud = 'dummy'
            taskSpec.cloud = None
        # site
        if 'site' in taskParamMap:
            self.siteName = taskParamMap['site']
            taskSpec.site = self.siteName
        else:
            # set dummy to force update
            taskSpec.site = 'dummy'
            taskSpec.site = None
        # nucleus
        if 'nucleus' in taskParamMap:
            taskSpec.nucleus = taskParamMap['nucleus']
        # preset some parameters for job cloning
        if 'useJobCloning' in taskParamMap:
            # set implicit parameters
            if 'nEventsPerWorker' not in taskParamMap:
                taskParamMap['nEventsPerWorker'] = 1
            if 'nSitesPerJob' not in taskParamMap:
                taskParamMap['nSitesPerJob'] = 2
            if 'nEsConsumers' not in taskParamMap:
                taskParamMap['nEsConsumers'] = taskParamMap['nSitesPerJob']
        # minimum granularity
        if 'minGranularity' in taskParamMap:
            taskParamMap['nEventsPerRange'] = taskParamMap['minGranularity']
        # event service flag
        if 'useJobCloning' in taskParamMap:
            taskSpec.eventService = 2
        elif 'nEventsPerWorker' in taskParamMap:
            taskSpec.eventService = 1
        else:
            taskSpec.eventService = 0
        # OS
        if 'osInfo' in taskParamMap:
            taskSpec.termCondition = taskParamMap['osInfo']
        # ttcr: requested time to completion
        if 'ttcrTimestamp' in taskParamMap:
            try:
                # get rid of the +00:00 timezone string and parse the timestamp
                taskSpec.ttcRequested = datetime.datetime.strptime(taskParamMap['ttcrTimestamp'].split('+')[0], '%Y-%m-%d %H:%M:%S.%f')
            except (IndexError, ValueError):
                pass
        # goal
        if 'goal' in taskParamMap:
            try:
                taskSpec.goal = int(float(taskParamMap['goal'])*10)
                if taskSpec.goal > 1000:
                    taskSpec.goal = None
            except Exception:
                pass
        # campaign
        if 'campaign' in taskParamMap:
            taskSpec.campaign = taskParamMap['campaign']
        # request type
        if 'requestType' in taskParamMap:
            taskSpec.requestType = taskParamMap['requestType']
        # image name
        if 'container_name' in taskParamMap:
            taskSpec.container_name = taskParamMap['container_name']
        self.taskSpec = taskSpec
        # set split rule
        if 'tgtNumEventsPerJob' in taskParamMap:
            # set nEventsPerJob not respect file boundaries when nFilesPerJob is not used
            if 'nFilesPerJob' not in taskParamMap:
                self.setSplitRule(None,taskParamMap['tgtNumEventsPerJob'],JediTaskSpec.splitRuleToken['nEventsPerJob'])
        self.setSplitRule(taskParamMap,'nFilesPerJob',     JediTaskSpec.splitRuleToken['nFilesPerJob'])
        self.setSplitRule(taskParamMap,'nEventsPerJob',    JediTaskSpec.splitRuleToken['nEventsPerJob'])
        self.setSplitRule(taskParamMap,'nGBPerJob',        JediTaskSpec.splitRuleToken['nGBPerJob'])
        self.setSplitRule(taskParamMap,'nMaxFilesPerJob',  JediTaskSpec.splitRuleToken['nMaxFilesPerJob'])
        self.setSplitRule(taskParamMap,'nEventsPerWorker', JediTaskSpec.splitRuleToken['nEventsPerWorker'])
        self.setSplitRule(taskParamMap,'disableAutoRetry', JediTaskSpec.splitRuleToken['disableAutoRetry'])
        self.setSplitRule(taskParamMap,'nEsConsumers',     JediTaskSpec.splitRuleToken['nEsConsumers'])
        self.setSplitRule(taskParamMap,'waitInput',        JediTaskSpec.splitRuleToken['waitInput'])
        self.setSplitRule(taskParamMap,'addNthFieldToLFN', JediTaskSpec.splitRuleToken['addNthFieldToLFN'])
        self.setSplitRule(taskParamMap,'scoutSuccessRate', JediTaskSpec.splitRuleToken['scoutSuccessRate'])
        self.setSplitRule(taskParamMap,'t1Weight',         JediTaskSpec.splitRuleToken['t1Weight'])
        self.setSplitRule(taskParamMap,'maxAttemptES',     JediTaskSpec.splitRuleToken['maxAttemptES'])
        self.setSplitRule(taskParamMap,'maxAttemptEsJob',  JediTaskSpec.splitRuleToken['maxAttemptEsJob'])
        self.setSplitRule(taskParamMap,'nSitesPerJob',     JediTaskSpec.splitRuleToken['nSitesPerJob'])
        self.setSplitRule(taskParamMap,'nEventsPerMergeJob',   JediTaskSpec.splitRuleToken['nEventsPerMergeJob'])
        self.setSplitRule(taskParamMap,'nFilesPerMergeJob',    JediTaskSpec.splitRuleToken['nFilesPerMergeJob'])
        self.setSplitRule(taskParamMap,'nGBPerMergeJob',       JediTaskSpec.splitRuleToken['nGBPerMergeJob'])
        self.setSplitRule(taskParamMap,'nMaxFilesPerMergeJob', JediTaskSpec.splitRuleToken['nMaxFilesPerMergeJob'])
        self.setSplitRule(taskParamMap,'maxWalltime', JediTaskSpec.splitRuleToken['maxWalltime'])
        self.setSplitRule(taskParamMap,'tgtMaxOutputForNG', JediTaskSpec.splitRuleToken['tgtMaxOutputForNG'])
        self.setSplitRule(taskParamMap, 'maxNumJobs', JediTaskSpec.splitRuleToken['maxNumJobs'])
        self.setSplitRule(taskParamMap, 'totNumJobs', JediTaskSpec.splitRuleToken['totNumJobs'])
        self.setSplitRule(taskParamMap, 'nChunksToWait', JediTaskSpec.splitRuleToken['nChunksToWait'])
        if 'forceStaged' in taskParamMap:
            taskParamMap['useLocalIO'] = taskParamMap['forceStaged']
        if 'useLocalIO' in taskParamMap:
            if taskParamMap['useLocalIO']:
                self.setSplitRule(None, 1, JediTaskSpec.splitRuleToken['useLocalIO'])
            else:
                self.setSplitRule(None, 0, JediTaskSpec.splitRuleToken['useLocalIO'])
        if 'nJumboJobs' in taskParamMap:
            self.setSplitRule(taskParamMap,'nJumboJobs',JediTaskSpec.splitRuleToken['nJumboJobs'])
            taskSpec.useJumbo = JediTaskSpec.enum_useJumbo['waiting']
            if 'maxJumboPerSite' in taskParamMap:
                self.setSplitRule(taskParamMap,'maxJumboPerSite',JediTaskSpec.splitRuleToken['maxJumboPerSite'])
        if 'minCpuEfficiency' in taskParamMap:
            self.setSplitRule(taskParamMap,'minCpuEfficiency',JediTaskSpec.splitRuleToken['minCpuEfficiency'])
        if 'loadXML' in taskParamMap:
            self.setSplitRule(None,3,JediTaskSpec.splitRuleToken['loadXML'])
            self.setSplitRule(None,4,JediTaskSpec.splitRuleToken['groupBoundaryID'])
        if 'pfnList' in taskParamMap:
            self.setSplitRule(None,1,JediTaskSpec.splitRuleToken['pfnList'])
        if 'noWaitParent' in taskParamMap and taskParamMap['noWaitParent'] is True:
            self.setSplitRule(None,1,JediTaskSpec.splitRuleToken['noWaitParent'])
        if 'respectLB' in taskParamMap:
            self.setSplitRule(None,1,JediTaskSpec.splitRuleToken['respectLB'])
        if 'releasePerLB' in taskParamMap:
                self.setSplitRule(None, 1, JediTaskSpec.splitRuleToken['releasePerLB'])
        if 'orderByLB' in taskParamMap:
            self.setSplitRule(None,1,JediTaskSpec.splitRuleToken['orderByLB'])
        if 'respectSplitRule' in taskParamMap:
            self.setSplitRule(None,1,JediTaskSpec.splitRuleToken['respectSplitRule'])
        if 'reuseSecOnDemand' in taskParamMap:
            self.setSplitRule(None,1,JediTaskSpec.splitRuleToken['reuseSecOnDemand'])
        if 'ddmBackEnd' in taskParamMap:
            self.taskSpec.setDdmBackEnd(taskParamMap['ddmBackEnd'])
        if 'disableReassign' in taskParamMap:
            self.setSplitRule(None,1,JediTaskSpec.splitRuleToken['disableReassign'])
        if 'allowPartialFinish' in taskParamMap:
            self.setSplitRule(None,1,JediTaskSpec.splitRuleToken['allowPartialFinish'])
        if 'useExhausted' in taskParamMap:
            self.setSplitRule(None,1,JediTaskSpec.splitRuleToken['useExhausted'])
        if 'useRealNumEvents' in taskParamMap:
            self.setSplitRule(None,1,JediTaskSpec.splitRuleToken['useRealNumEvents'])
        if 'ipConnectivity' in taskParamMap:
            self.taskSpec.setIpConnectivity(taskParamMap['ipConnectivity'])
        if 'altStageOut' in taskParamMap:
            self.taskSpec.setAltStageOut(taskParamMap['altStageOut'])
        if 'allowInputLAN' in taskParamMap:
            self.taskSpec.setAllowInputLAN(taskParamMap['allowInputLAN'])
        if 'runUntilClosed' in taskParamMap:
            self.setSplitRule(None,1,JediTaskSpec.splitRuleToken['runUntilClosed'])
        if 'stayOutputOnSite' in taskParamMap:
            self.setSplitRule(None,1,JediTaskSpec.splitRuleToken['stayOutputOnSite'])
        if 'useJobCloning' in taskParamMap:
            scValue = EventServiceUtils.getJobCloningValue(taskParamMap['useJobCloning'])
            self.setSplitRule(None,scValue,JediTaskSpec.splitRuleToken['useJobCloning'])
        if 'failWhenGoalUnreached' in taskParamMap and taskParamMap['failWhenGoalUnreached'] is True:
            self.setSplitRule(None,1,JediTaskSpec.splitRuleToken['failGoalUnreached'])
        if 'switchEStoNormal' in taskParamMap:
            self.setSplitRule(None,1,JediTaskSpec.splitRuleToken['switchEStoNormal'])
        if 'nEventsPerRange' in taskParamMap:
            self.setSplitRule(taskParamMap,'nEventsPerRange',JediTaskSpec.splitRuleToken['dynamicNumEvents'])
        if 'allowInputWAN' in taskParamMap and taskParamMap['allowInputWAN'] is True:
            self.setSplitRule(None,1,JediTaskSpec.splitRuleToken['allowInputWAN'])
        if 'putLogToOS' in taskParamMap and taskParamMap['putLogToOS'] is True:
            self.setSplitRule(None,1,JediTaskSpec.splitRuleToken['putLogToOS'])
        if 'mergeEsOnOS' in taskParamMap and taskParamMap['mergeEsOnOS'] is True:
            self.setSplitRule(None,1,JediTaskSpec.splitRuleToken['mergeEsOnOS'])
        if 'writeInputToFile' in taskParamMap and taskParamMap['writeInputToFile'] is True:
            self.setSplitRule(None,1,JediTaskSpec.splitRuleToken['writeInputToFile'])
        if 'useFileAsSourceLFN' in taskParamMap and taskParamMap['useFileAsSourceLFN'] is True:
            self.setSplitRule(None,1,JediTaskSpec.splitRuleToken['useFileAsSourceLFN'])
        if 'ignoreMissingInDS' in taskParamMap and taskParamMap['ignoreMissingInDS'] is True:
            self.setSplitRule(None,1,JediTaskSpec.splitRuleToken['ignoreMissingInDS'])
        if 'noExecStrCnv' in taskParamMap and taskParamMap['noExecStrCnv'] is True:
            self.setSplitRule(None,1,JediTaskSpec.splitRuleToken['noExecStrCnv'])
        if 'inFilePosEvtNum' in taskParamMap and taskParamMap['inFilePosEvtNum'] is True:
            self.setSplitRule(None,1,JediTaskSpec.splitRuleToken['inFilePosEvtNum'])
        if self.taskSpec.useEventService() and not taskSpec.useJobCloning():
            if 'registerEsFiles' in taskParamMap and taskParamMap['registerEsFiles'] is True:
                self.setSplitRule(None,1,JediTaskSpec.splitRuleToken['registerEsFiles'])
        if 'disableAutoFinish' in taskParamMap and taskParamMap['disableAutoFinish'] is True:
            self.setSplitRule(None,1,JediTaskSpec.splitRuleToken['disableAutoFinish'])
        if 'resurrectConsumers' in taskParamMap and taskParamMap['resurrectConsumers'] is True:
            self.setSplitRule(None,1,JediTaskSpec.splitRuleToken['resurrectConsumers'])
        if 'usePrefetcher' in taskParamMap and taskParamMap['usePrefetcher'] is True:
            self.setSplitRule(None,1,JediTaskSpec.splitRuleToken['usePrefetcher'])
        if 'notDiscardEvents' in taskParamMap and taskParamMap['notDiscardEvents'] is True:
            self.setSplitRule(None,1,JediTaskSpec.splitRuleToken['notDiscardEvents'])
        if 'decAttOnFailedES' in taskParamMap and taskParamMap['decAttOnFailedES'] is True:
            self.setSplitRule(None, 1, JediTaskSpec.splitRuleToken['decAttOnFailedES'])
        if 'useZipToPin' in taskParamMap and taskParamMap['useZipToPin'] is True:
            self.setSplitRule(None, 1, JediTaskSpec.splitRuleToken['useZipToPin'])
        if 'osMatching' in taskParamMap and taskParamMap['osMatching'] is True:
            self.setSplitRule(None, 1, JediTaskSpec.splitRuleToken['osMatching'])
        if 'multiStepExec' in taskParamMap:
            self.setSplitRule(None, 1, JediTaskSpec.splitRuleToken['multiStepExec'])
        if 'onlyTagsForFC' in taskParamMap:
            self.setSplitRule(None, 1, JediTaskSpec.splitRuleToken['onlyTagsForFC'])
        if 'segmentedWork' in taskParamMap and 'segmentSpecs' in taskParamMap:
            self.setSplitRule(None, 1, JediTaskSpec.splitRuleToken['segmentedWork'])
        if 'avoidVP' in taskParamMap:
                self.setSplitRule(None, 1, JediTaskSpec.splitRuleToken['avoidVP'])
        if 'inputPreStaging' in taskParamMap and taskParamMap['inputPreStaging'] is True:
            self.setSplitRule(None, JediTaskSpec.enum_inputPreStaging['use'],
                              JediTaskSpec.splitRuleToken['inputPreStaging'])
        if 'hpoWorkflow' in taskParamMap and taskParamMap['hpoWorkflow'] is True and 'hpoRequestData' in taskParamMap:
            self.setSplitRule(None, 1, JediTaskSpec.splitRuleToken['hpoWorkflow'])
        if 'noLoopingCheck' in taskParamMap and taskParamMap['noLoopingCheck']:
            self.setSplitRule(None, 1, JediTaskSpec.splitRuleToken['noLoopingCheck'])
        if 'encJobParams' in taskParamMap and taskParamMap['encJobParams']:
            self.setSplitRule(None, 1, JediTaskSpec.splitRuleToken['encJobParams'])
        if 'useSecrets' in taskParamMap and taskParamMap['useSecrets']:
            self.setSplitRule(None, 1, JediTaskSpec.splitRuleToken['useSecrets'])
        if 'debugMode' in taskParamMap and taskParamMap['debugMode']:
            self.setSplitRule(None, 1, JediTaskSpec.splitRuleToken['debugMode'])
        if 'pushStatusChanges' in taskParamMap and taskParamMap['pushStatusChanges']:
            self.setSplitRule(None, 1, JediTaskSpec.splitRuleToken['pushStatusChanges'])
        if 'maxCoreCount' in taskParamMap:
            self.setSplitRule(taskParamMap, 'maxCoreCount', JediTaskSpec.splitRuleToken['maxCoreCount'])
        if 'cloudAsVO' in taskParamMap:
            self.setSplitRule(None, 1, JediTaskSpec.splitRuleToken['cloudAsVO'])
        # work queue
        workQueue = None
        if 'workQueueName' in taskParamMap:
            # work queue is specified
            workQueue = workQueueMapper.getQueueByName(taskSpec.vo, taskSpec.prodSourceLabel, taskParamMap['workQueueName'])
        if workQueue is None:
            # get work queue based on task attributes
            workQueue,tmpStr = workQueueMapper.getQueueWithSelParams(taskSpec.vo,
                                                                     taskSpec.prodSourceLabel,
                                                                     prodSourceLabel=taskSpec.prodSourceLabel,
                                                                     processingType=taskSpec.processingType,
                                                                     workingGroup=taskSpec.workingGroup,
                                                                     coreCount=taskSpec.coreCount,
                                                                     site=taskSpec.site,
                                                                     eventService=taskSpec.eventService,
                                                                     splitRule=taskSpec.splitRule,
                                                                     campaign=taskSpec.campaign)
        if workQueue is None:
            errStr  = 'workqueue is undefined for vo={0} label={1} '.format(taskSpec.vo,taskSpec.prodSourceLabel)
            errStr += 'processingType={0} workingGroup={1} coreCount={2} eventService={3} '.format(taskSpec.processingType,
                                                                                                   taskSpec.workingGroup,
                                                                                                   taskSpec.coreCount,
                                                                                                   taskSpec.eventService)
            errStr += 'splitRule={0} campaign={1}'.format(taskSpec.splitRule,taskSpec.campaign)
            raise RuntimeError(errStr)
        self.taskSpec.workQueue_ID = workQueue.queue_id

        # Initialize the global share
        gshare = 'Undefined'
        if 'gshare' in taskParamMap and self.taskBufferIF.is_valid_share(taskParamMap['gshare']):
            # work queue is specified
            gshare = taskParamMap['gshare']
        else:
            # get share based on definition
            gshare = self.taskBufferIF.get_share_for_task(self.taskSpec)
            if gshare is None:
                gshare = 'Undefined'  # Should not happen. Undefined is set when no share is found
                # errStr  = 'share is undefined for vo={0} label={1} '.format(taskSpec.vo,taskSpec.prodSourceLabel)
                # errStr += 'workingGroup={0} campaign={1} '.format(taskSpec.workingGroup, taskSpec.campaign)
                # raise RuntimeError,errStr
        self.taskSpec.gshare = gshare

        # Initialize the resource type
        try:
            self.taskSpec.resource_type = self.taskBufferIF.get_resource_type_task(self.taskSpec)
        except Exception:
            self.taskSpec.resource_type = 'Undefined'

        # return
        return



    # basic refinement procedure
    def doBasicRefine(self,taskParamMap):
        # get input/output/log dataset specs
        nIn  = 0
        nOutMap = {}
        if 'log' not in taskParamMap:
            itemList = taskParamMap['jobParameters']
        elif isinstance(taskParamMap['log'],dict):
            itemList = taskParamMap['jobParameters'] + [taskParamMap['log']]
        else:
            itemList = taskParamMap['jobParameters'] + taskParamMap['log']
        if 'log_merge' in taskParamMap:
            itemList += [taskParamMap['log_merge']]
        # pseudo input
        if 'noInput' in taskParamMap and taskParamMap['noInput'] is True:
            tmpItem = {}
            tmpItem['type']       = 'template'
            tmpItem['value']      = ''
            tmpItem['dataset']    = 'pseudo_dataset'
            tmpItem['param_type'] = 'pseudo_input'
            itemList = [tmpItem] + itemList
        # random seed
        if RefinerUtils.useRandomSeed(taskParamMap):
            tmpItem = {}
            tmpItem['type']       = 'template'
            tmpItem['value']      = ''
            tmpItem['dataset']    = 'RNDMSEED'
            tmpItem['param_type'] = 'random_seed'
            itemList.append(tmpItem)
        # loop over all items
        allDsList = []
        for tmpItem in itemList:
            # look for datasets
            if tmpItem['type'] == 'template' and 'dataset' in tmpItem:
                # avoid duplication
                if tmpItem['dataset'] not in allDsList:
                    allDsList.append(tmpItem['dataset'])
                else:
                    continue
                datasetSpec = JediDatasetSpec()
                datasetSpec.datasetName = tmpItem['dataset']
                datasetSpec.jediTaskID = self.taskSpec.jediTaskID
                datasetSpec.type = tmpItem['param_type']
                if 'container' in tmpItem:
                    datasetSpec.containerName = tmpItem['container']
                if 'token' in tmpItem:
                    datasetSpec.storageToken = tmpItem['token']
                if 'destination' in tmpItem:
                    datasetSpec.destination = tmpItem['destination']
                if 'attributes' in tmpItem:
                    datasetSpec.setDatasetAttribute(tmpItem['attributes'])
                if 'ratio' in tmpItem:
                    datasetSpec.setDatasetAttribute('ratio={0}'.format(tmpItem['ratio']))
                if 'eventRatio' in tmpItem:
                    datasetSpec.setEventRatio(tmpItem['eventRatio'])
                if 'check' in tmpItem:
                    datasetSpec.setDatasetAttribute('cc')
                if 'usedup' in tmpItem:
                    datasetSpec.setDatasetAttribute('ud')
                if 'random' in tmpItem:
                    datasetSpec.setDatasetAttribute('rd')
                if 'reusable' in tmpItem:
                    datasetSpec.setDatasetAttribute('ru')
                if 'indexConsistent' in tmpItem:
                    datasetSpec.setDatasetAttributeWithLabel('indexConsistent')
                if 'mergeOnly' in tmpItem:
                    datasetSpec.setDatasetAttributeWithLabel('mergeOnly')
                if 'offset' in tmpItem:
                    datasetSpec.setOffset(tmpItem['offset'])
                if 'allowNoOutput' in tmpItem:
                    datasetSpec.allowNoOutput()
                if 'nFilesPerJob' in tmpItem:
                    datasetSpec.setNumFilesPerJob(tmpItem['nFilesPerJob'])
                if 'num_records' in tmpItem:
                    datasetSpec.setNumRecords(tmpItem['num_records'])
                if 'transient' in tmpItem:
                    datasetSpec.setTransient(tmpItem['transient'])
                if 'pseudo' in tmpItem:
                    datasetSpec.setPseudo()
                datasetSpec.vo = self.taskSpec.vo
                datasetSpec.nFiles = 0
                datasetSpec.nFilesUsed = 0
                datasetSpec.nFilesFinished = 0
                datasetSpec.nFilesFailed = 0
                datasetSpec.nFilesOnHold = 0
                datasetSpec.nFilesWaiting = 0
                datasetSpec.nEvents = 0
                datasetSpec.nEventsUsed = 0
                datasetSpec.nEventsToBeUsed = 0
                datasetSpec.status = 'defined'
                if datasetSpec.type in JediDatasetSpec.getInputTypes() + ['random_seed']:
                    datasetSpec.streamName = RefinerUtils.extractStreamName(tmpItem['value'])
                    if 'expandedList' not in tmpItem:
                        tmpItem['expandedList'] = []
                    # dataset names could be comma-concatenated
                    datasetNameList = datasetSpec.datasetName.split(',')
                    # datasets could be added by incexec
                    incexecDS = 'dsFor{0}'.format(datasetSpec.streamName)
                    # remove /XYZ
                    incexecDS = incexecDS.split('/')[0]
                    if incexecDS in taskParamMap:
                        for tmpDatasetName in taskParamMap[incexecDS].split(','):
                            if tmpDatasetName not in datasetNameList:
                                datasetNameList.append(tmpDatasetName)
                    # loop over all dataset names
                    inDatasetSpecList = []
                    for datasetName in datasetNameList:
                        # skip empty
                        if datasetName == '':
                            continue
                        # expand
                        if datasetSpec.isPseudo() or datasetSpec.type in ['random_seed'] or datasetName == 'DBR_LATEST':
                            # pseudo input
                            tmpDatasetNameList = [datasetName]
                            if self.taskSpec.is_work_segmented():
                                tmpDatasetNameList *= len(taskParamMap['segmentSpecs'])
                        else:
                            tmpIF = self.ddmIF.getInterface(self.taskSpec.vo, self.taskSpec.cloud)
                            if not tmpIF:
                                tmpDatasetNameList = []
                            else:
                                if 'expand' in tmpItem and tmpItem['expand'] is True:
                                    # expand dataset container
                                    tmpDatasetNameList = tmpIF.expandContainer(datasetName)
                                else:
                                    # normal dataset name
                                    tmpDatasetNameList = tmpIF.listDatasets(datasetName)
                        i_element = 0
                        for elementDatasetName in tmpDatasetNameList:
                            if nIn > 0 or elementDatasetName not in tmpItem['expandedList'] or \
                                    self.taskSpec.is_work_segmented():
                                tmpItem['expandedList'].append(elementDatasetName)
                                inDatasetSpec = copy.copy(datasetSpec)
                                inDatasetSpec.datasetName = elementDatasetName
                                if nIn > 0 or not self.taskSpec.is_hpo_workflow():
                                    inDatasetSpec.containerName = datasetName
                                else:
                                    if self.taskSpec.is_work_segmented():
                                        inDatasetSpec.containerName = "{}/{}".format(
                                            taskParamMap['segmentSpecs'][i_element]['name'],
                                            taskParamMap['segmentSpecs'][i_element]['id']
                                        )
                                    else:
                                        inDatasetSpec.containerName = "None/None"
                                inDatasetSpecList.append(inDatasetSpec)
                            i_element += 1
                    # empty input
                    if inDatasetSpecList == [] and self.oldTaskStatus != 'rerefine':
                        errStr = 'doBasicRefine : unknown input dataset "{0}"'.format(datasetSpec.datasetName)
                        self.taskSpec.setErrDiag(errStr)
                        if datasetSpec.datasetName not in self.unknownDatasetList:
                            self.unknownDatasetList.append(datasetSpec.datasetName)
                        raise JediException.UnknownDatasetError(errStr)
                    # set master flag
                    for inDatasetSpec in inDatasetSpecList:
                        if nIn == 0:
                            # master
                            self.inMasterDatasetSpec.append(inDatasetSpec)
                        else:
                            # secondary
                            self.inSecDatasetSpecList.append(inDatasetSpec)
                    nIn += 1
                    continue
                if datasetSpec.type in ['output','log']:
                    if datasetSpec.type not in nOutMap:
                        nOutMap[datasetSpec.type] = 0
                    # make stream name
                    if not datasetSpec.is_merge_only():
                        datasetSpec.streamName = "{0}{1}".format(datasetSpec.type.upper(),nOutMap[datasetSpec.type])
                    else:
                        datasetSpec.streamName = 'LOG_MERGE'
                    nOutMap[datasetSpec.type] += 1
                    # set attribute for event service
                    if self.taskSpec.useEventService() and 'objectStore' in taskParamMap and datasetSpec.type in ['output']:
                        datasetSpec.setObjectStore(taskParamMap['objectStore'])
                    # extract output filename template and change the value field
                    outFileTemplate,tmpItem['value'] = RefinerUtils.extractReplaceOutFileTemplate(tmpItem['value'],
                                                                                                  datasetSpec.streamName)
                    # make output template
                    if outFileTemplate is not None:
                        if 'offset' in tmpItem:
                            offsetVal = 1 + tmpItem['offset']
                        else:
                            offsetVal = 1
                        outTemplateMap = {'jediTaskID' : self.taskSpec.jediTaskID,
                                          'serialNr' : offsetVal,
                                          'streamName' : datasetSpec.streamName,
                                          'filenameTemplate' : outFileTemplate,
                                          'outtype' : datasetSpec.type,
                                          }
                        if datasetSpec.outputMapKey() in self.outputTemplateMap:
                            # multiple files are associated to the same output datasets
                            self.outputTemplateMap[datasetSpec.outputMapKey()].append(outTemplateMap)
                            # don't insert the same output dataset
                            continue
                        self.outputTemplateMap[datasetSpec.outputMapKey()] = [outTemplateMap]
                    # append
                    self.outDatasetSpecList.append(datasetSpec)
                    # used only in merge
                    if datasetSpec.is_merge_only():
                        continue
                    # make unmerged dataset
                    if 'mergeOutput' in taskParamMap and taskParamMap['mergeOutput'] is True:
                        umDatasetSpec = JediDatasetSpec()
                        umDatasetSpec.datasetName = 'panda.um.' + datasetSpec.datasetName
                        umDatasetSpec.jediTaskID = self.taskSpec.jediTaskID
                        umDatasetSpec.storageToken = 'TOMERGE'
                        umDatasetSpec.vo = datasetSpec.vo
                        umDatasetSpec.type = "tmpl_trn_" + datasetSpec.type
                        umDatasetSpec.nFiles = 0
                        umDatasetSpec.nFilesUsed = 0
                        umDatasetSpec.nFilesToBeUsed = 0
                        umDatasetSpec.nFilesFinished = 0
                        umDatasetSpec.nFilesFailed = 0
                        umDatasetSpec.nFilesOnHold = 0
                        umDatasetSpec.status = 'defined'
                        umDatasetSpec.streamName = datasetSpec.streamName
                        if datasetSpec.isAllowedNoOutput():
                            umDatasetSpec.allowNoOutput()
                        # ratio
                        if datasetSpec.getRatioToMaster() > 1:
                            umDatasetSpec.setDatasetAttribute('ratio={0}'.format(datasetSpec.getRatioToMaster()))
                        # make unmerged output template
                        if outFileTemplate is not None:
                            umOutTemplateMap = {'jediTaskID' : self.taskSpec.jediTaskID,
                                                'serialNr' : 1,
                                                'streamName' : umDatasetSpec.streamName,
                                                'outtype' : datasetSpec.type,
                                                }
                            # append temporary name
                            if 'umNameAtEnd' in taskParamMap and taskParamMap['umNameAtEnd'] is True:
                                # append temporary name at the end
                                umOutTemplateMap['filenameTemplate'] = outFileTemplate + '.panda.um'
                            else:
                                umOutTemplateMap['filenameTemplate'] = 'panda.um.' + outFileTemplate
                            if umDatasetSpec.outputMapKey() in self.outputTemplateMap:
                                # multiple files are associated to the same output datasets
                                self.outputTemplateMap[umDatasetSpec.outputMapKey()].append(umOutTemplateMap)
                                # don't insert the same output dataset
                                continue
                            self.outputTemplateMap[umDatasetSpec.outputMapKey()] = [umOutTemplateMap]
                        # use log as master for merging
                        if datasetSpec.type == 'log':
                            self.unmergeMasterDatasetSpec[datasetSpec.outputMapKey()] = umDatasetSpec
                        else:
                            # append
                            self.unmergeDatasetSpecMap[datasetSpec.outputMapKey()] = umDatasetSpec
        # set attributes for merging
        if 'mergeOutput' in taskParamMap and taskParamMap['mergeOutput'] is True:
            self.setSplitRule(None,1,JediTaskSpec.splitRuleToken['mergeOutput'])
        # make job parameters
        rndmSeedOffset = None
        firstEventOffset = None
        jobParameters = ''
        for tmpItem in taskParamMap['jobParameters']:
            if 'value' in tmpItem:
                # hidden parameter
                if 'hidden' in tmpItem and tmpItem['hidden'] is True:
                    continue
                # add tags for ES-only parameters
                esOnly = False
                if 'es_only' in tmpItem and tmpItem['es_only'] is True:
                    esOnly = True
                if esOnly:
                    jobParameters += '<PANDA_ES_ONLY>'
                jobParameters += '{0}'.format(tmpItem['value'])
                if esOnly:
                    jobParameters += '</PANDA_ES_ONLY>'
                # padding
                if 'padding' in tmpItem and tmpItem['padding'] is False:
                    pass
                else:
                    jobParameters += ' '
                # get offset for random seed and first event
                if tmpItem['type'] == 'template' and tmpItem['param_type'] == 'number':
                    if '${RNDMSEED}' in tmpItem['value']:
                        if 'offset' in tmpItem:
                            rndmSeedOffset = tmpItem['offset']
                        else:
                            rndmSeedOffset = 0
                    elif '${FIRSTEVENT}' in tmpItem['value']:
                        if 'offset' in tmpItem:
                            firstEventOffset = tmpItem['offset']
        jobParameters = jobParameters[:-1]
        # append parameters for event service merging if necessary
        esmergeParams = self.getParamsForEventServiceMerging(taskParamMap)
        if esmergeParams is not None:
            jobParameters += esmergeParams
        self.setJobParamsTemplate(jobParameters)
        # set random seed offset
        if rndmSeedOffset is not None:
            self.setSplitRule(None,rndmSeedOffset,JediTaskSpec.splitRuleToken['randomSeed'])
        if firstEventOffset is not None:
            self.setSplitRule(None,firstEventOffset,JediTaskSpec.splitRuleToken['firstEvent'])
        # send HPO request
        if self.taskSpec.is_hpo_workflow():
            try:
                data = copy.copy(taskParamMap['hpoRequestData'])
                data['workload_id'] = self.taskSpec.jediTaskID
                data['is_pseudo_input'] = True
                req = {
                    'requester': 'panda',
                    'request_type': idds.common.constants.RequestType.HyperParameterOpt,
                    'transform_tag': idds.common.constants.RequestType.HyperParameterOpt.value,
                    'status': idds.common.constants.RequestStatus.New,
                    'priority': 0,
                    'lifetime': 30,
                    'request_metadata': data,
                }
                c = iDDS_Client(idds.common.utils.get_rest_host())
                self.tmpLog.debug('req {0}'.format(str(req)))
                ret = c.add_request(**req)
                self.tmpLog.debug('got requestID={0}'.format(str(ret)))
            except Exception as e:
                errStr = 'iDDS failed with {0}'.format(str(e))
                raise JediException.ExternalTempError(errStr)
        # return
        return



    # replace placeholder with dict provided by prepro job
    def replacePlaceHolders(self,paramItem,placeHolderName,newValue):
        if isinstance(paramItem, dict):
            # loop over all dict params
            for tmpParName,tmpParVal in iteritems(paramItem):
                if tmpParVal == placeHolderName:
                    # replace placeholder
                    paramItem[tmpParName] = newValue
                elif isinstance(tmpParVal, dict) or \
                        isinstance(tmpParVal, list):
                    # recursive execution
                    self.replacePlaceHolders(tmpParVal,placeHolderName,newValue)
        elif isinstance(paramItem, list):
            # loop over all list items
            for tmpItem in paramItem:
                self.replacePlaceHolders(tmpItem,placeHolderName,newValue)



    # refinement procedure for preprocessing
    def doPreProRefine(self,taskParamMap):
        # no preprocessing
        if 'preproSpec' not in taskParamMap:
            return None,taskParamMap
        # already preprocessed
        if self.taskSpec.checkPreProcessed():
            # get replaced task params
            tmpStat,tmpJsonStr = self.taskBufferIF.getPreprocessMetadata_JEDI(self.taskSpec.jediTaskID)
            try:
                # replace placeholders
                replaceParams = RefinerUtils.decodeJSON(tmpJsonStr)
                self.tmpLog.debug("replace placeholders with "+str(replaceParams))
                for tmpKey,tmpVal in iteritems(replaceParams):
                    self.replacePlaceHolders(taskParamMap,tmpKey,tmpVal)
            except Exception:
                errtype,errvalue = sys.exc_info()[:2]
                self.tmpLog.error('{0} failed to get additional task params with {1}:{2}'.format(self.__class__.__name__,
                                                                                                 errtype.__name__,errvalue))
                return False,taskParamMap
            # succeeded
            self.updatedTaskParams = taskParamMap
            return None,taskParamMap
        # make dummy dataset to keep track of preprocessing
        datasetSpec = JediDatasetSpec()
        datasetSpec.datasetName = 'panda.pp.in.{0}.{1}'.format(uuid.uuid4(),self.taskSpec.jediTaskID)
        datasetSpec.jediTaskID = self.taskSpec.jediTaskID
        datasetSpec.type = 'pp_input'
        datasetSpec.vo = self.taskSpec.vo
        datasetSpec.nFiles = 1
        datasetSpec.nFilesUsed = 0
        datasetSpec.nFilesToBeUsed = 1
        datasetSpec.nFilesFinished = 0
        datasetSpec.nFilesFailed = 0
        datasetSpec.nFilesOnHold = 0
        datasetSpec.status = 'ready'
        self.inMasterDatasetSpec.append(datasetSpec)
        # make file
        fileSpec = JediFileSpec()
        fileSpec.jediTaskID   = datasetSpec.jediTaskID
        fileSpec.type         = datasetSpec.type
        fileSpec.status       = 'ready'
        fileSpec.lfn          = 'pseudo_lfn'
        fileSpec.attemptNr    = 0
        fileSpec.maxAttempt   = 3
        fileSpec.keepTrack    = 1
        datasetSpec.addFile(fileSpec)
        # make log dataset
        logDatasetSpec = JediDatasetSpec()
        logDatasetSpec.datasetName = 'panda.pp.log.{0}.{1}'.format(uuid.uuid4(),self.taskSpec.jediTaskID)
        logDatasetSpec.jediTaskID = self.taskSpec.jediTaskID
        logDatasetSpec.type = 'tmpl_pp_log'
        logDatasetSpec.streamName = 'PP_LOG'
        logDatasetSpec.vo = self.taskSpec.vo
        logDatasetSpec.nFiles = 0
        logDatasetSpec.nFilesUsed = 0
        logDatasetSpec.nFilesToBeUsed = 0
        logDatasetSpec.nFilesFinished = 0
        logDatasetSpec.nFilesFailed = 0
        logDatasetSpec.nFilesOnHold = 0
        logDatasetSpec.status = 'defined'
        self.outDatasetSpecList.append(logDatasetSpec)
        # make output template for log
        outTemplateMap = {'jediTaskID' : self.taskSpec.jediTaskID,
                          'serialNr' : 1,
                          'streamName' : logDatasetSpec.streamName,
                          'filenameTemplate' : "{0}._${{SN}}.log.tgz".format(logDatasetSpec.datasetName),
                          'outtype' : re.sub('^tmpl_','',logDatasetSpec.type),
                          }
        self.outputTemplateMap[logDatasetSpec.outputMapKey()] = [outTemplateMap]
        # set split rule to use preprocessing
        self.taskSpec.setPrePro()
        # set task status
        self.taskSpec.status = 'topreprocess'
        # return
        return True,taskParamMap



    # set split rule
    def setSplitRule(self,taskParamMap,keyName,valName):
        if taskParamMap is not None:
            if keyName not in taskParamMap:
                return
            tmpStr = '{0}={1}'.format(valName,taskParamMap[keyName])
        else:
            tmpStr = '{0}={1}'.format(valName,keyName)
        if self.taskSpec.splitRule in [None,'']:
            self.taskSpec.splitRule = tmpStr
        else:
            tmpMatch = re.search(valName+'=(-*\d+)(,-*\d+)*',self.taskSpec.splitRule)
            if tmpMatch is None:
                # append
                self.taskSpec.splitRule += ',{0}'.format(tmpStr)
            else:
                # replace
                self.taskSpec.splitRule = self.taskSpec.splitRule.replace(tmpMatch.group(0), tmpStr)
        return



    # get parameters for event service merging
    def getParamsForEventServiceMerging(self,taskParamMap):
        # no event service
        if not self.taskSpec.useEventService():
            return None
        # extract parameters
        transPath = 'UnDefined'
        jobParameters = 'UnDefined'
        if 'esmergeSpec' in taskParamMap:
            if 'transPath' in taskParamMap['esmergeSpec']:
                transPath = taskParamMap['esmergeSpec']['transPath']
            if 'jobParameters' in taskParamMap['esmergeSpec']:
                jobParameters = taskParamMap['esmergeSpec']['jobParameters']
        # return
        return '<PANDA_ESMERGE_TRF>'+transPath+'</PANDA_ESMERGE_TRF>'+'<PANDA_ESMERGE_JOBP>'+jobParameters+'</PANDA_ESMERGE_JOBP>'



Interaction.installSC(TaskRefinerBase)



import re
import os
import copy
import math
from pandaserver.taskbuffer import JobUtils

try:
    long()
except Exception:
    long = int


# get effective file size
def getEffectiveFileSize(fsize,startEvent,endEvent,nEvents):
    inMB = 1024 * 1024
    if fsize in [None,0]:
        # use dummy size for pseudo input
        effectiveFsize = inMB
    elif nEvents is not None and startEvent is not None and endEvent is not None:
        # take event range into account
        effectiveFsize = long(float(fsize)*float(endEvent-startEvent+1)/float(nEvents))
    else:
        effectiveFsize = fsize
    # use dummy size if input is too small
    if effectiveFsize == 0:
        effectiveFsize = inMB
    # in MB
    effectiveFsize = float(effectiveFsize) / inMB
    # return
    return effectiveFsize



# get effective number of events
def getEffectiveNumEvents(startEvent,endEvent,nEvents):
    if endEvent is not None and startEvent is not None:
        evtCounts = endEvent-startEvent+1
        if evtCounts > 0:
            return evtCounts
        return 1
    if nEvents is not None:
        return nEvents
    return 1



# get memory usage
def getMemoryUsage():
    try:
        t = open('/proc/{0}/status'.format(os.getpid()))
        v = t.read()
        t.close()
        value = 0
        for line in v.split('\n'):
            if line.startswith('VmRSS'):
                items = line.split()
                value = int(items[1])
                if items[2] in ['kB','KB']:
                    value /= 1024
                elif items[2] in ['mB','MB']:
                    pass
                break
        return value
    except Exception:
        return None



# check process
def checkProcess(pid):
    return os.path.exists('/proc/{0}/status'.format(pid))



# offset for walltime
wallTimeOffset = 10*60

# add offset to walltime
def addOffsetToWalltime(oldWalltime):
    if oldWalltime > 0:
        # add offset of 10min
        oldWalltime += wallTimeOffset
    return oldWalltime


# reduce offset from walltime
def reduceOffsetFromWalltime(oldWalltime):
    if oldWalltime > 0:
        # add offset of 10min
        oldWalltime -= wallTimeOffset
        if oldWalltime < 0:
            oldWalltime = 0
    return oldWalltime



# get config param for vo and prodSourceLabel
def getConfigParam(configStr,vo,sourceLabel):
    try:
        for tmpConfigStr in configStr.split(','):
            items = configStr.split(':')
            vos          = items[0].split('|')
            sourceLabels = items[1].split('|')
            if vo not in ['','any'] and \
                    vo not in vos and \
                    None not in vos and \
                    'any' not in vos and \
                    '' not in vos:
                continue
            if sourceLabel not in ['','any'] and \
                    sourceLabel not in sourceLabels and \
                    None not in sourceLabels and \
                    'any' not in sourceLabels and \
                    '' not in sourceLabels:
                continue
            return ','.join(items[2:])
    except Exception:
        pass
    return None


# get percentile until numpy 1.5.X becomes available
def percentile(inList,percent,idMap):
    inList = copy.copy(inList)
    inList.sort()
    k = (len(inList)-1) * float(percent)/100
    f = math.floor(k)
    c = math.ceil(k)
    if f == c:
        retVal = inList[int(f)]
        return retVal,[retVal]
    val0 = inList[int(f)]
    val1 = inList[int(c)]
    d0 = val0 * (c-k)
    d1 = val1 * (k-f)
    retVal = d0+d1
    return retVal,[val0,val1]


# get min ram count for job
def getJobMinRamCount(taskSpec, inputChunk, siteSpec, coreCount):
    minRamCount = inputChunk.getMaxRamCount()
    if inputChunk.isMerging:
        minRamUnit = 'MB'
    else:
        minRamUnit = taskSpec.ramUnit
        if minRamUnit in [None,'','NULL']:
            minRamUnit   = 'MB'
        if taskSpec.ramPerCore():
            minRamCount *= coreCount
            minRamCount += taskSpec.baseRamCount
            minRamUnit = re.sub('PerCore.*$', '', minRamUnit)
    # round up with chunks
    minRamCount = JobUtils.compensate_ram_count(minRamCount)
    return minRamCount, minRamUnit



# get max walltime and cpu count
def getJobMaxWalltime(taskSpec, inputChunk, totalMasterEvents, jobSpec, siteSpec):
    try:
        if not taskSpec.getCpuTime():
            jobSpec.maxWalltime = siteSpec.maxtime
            jobSpec.maxCpuCount = siteSpec.maxtime
        else:
            jobSpec.maxWalltime = taskSpec.getCpuTime()
            if jobSpec.maxWalltime is not None and jobSpec.maxWalltime > 0:
                jobSpec.maxWalltime *= totalMasterEvents
                if siteSpec.coreCount > 0:
                    jobSpec.maxWalltime /= float(siteSpec.coreCount)
                if siteSpec.corepower not in [0,None]:
                    jobSpec.maxWalltime /= siteSpec.corepower
            if taskSpec.cpuEfficiency not in [None,0]:
                jobSpec.maxWalltime /= (float(taskSpec.cpuEfficiency) / 100.0)
            if taskSpec.baseWalltime is not None:
                jobSpec.maxWalltime += taskSpec.baseWalltime
            jobSpec.maxWalltime = long(jobSpec.maxWalltime)
            if taskSpec.useHS06():
                jobSpec.maxCpuCount = jobSpec.maxWalltime
    except Exception:
        pass


# use direct IO for job
def use_direct_io_for_job(task_spec, site_spec, input_chunk):
    # not for merging
    if input_chunk and input_chunk.isMerging:
        return False
    # force copy-to-scratch
    if task_spec.useLocalIO():
        return False
    # always
    if site_spec.always_use_direct_io():
        return True
    # depends on task and site specs
    if task_spec.allowInputLAN() is not None and site_spec.isDirectIO():
        return True
    return False




from pandacommon.pandalogger import logger_utils

from pandajedi.jediconfig import jedi_config
from pandajedi.jedidaemons.utils import DaemonMaster


# logger
base_logger = logger_utils.setup_logger(__name__.split('.')[-1])


# launch
def launcher(taskBufferIF, ddmIF):
    tmp_log = logger_utils.make_logger(base_logger, method_name='launcher')
    tmp_log.debug('start')
    try:
        jedi_config.daemon.config
    except Exception as e:
        tmp_log.error('failed to read config json file; should not happen... {0}: {1}'.format(e.__class__.__name__, e))
        raise e
    # whether to run daemons
    if not getattr(jedi_config.daemon, 'enable', False):
        tmp_log.debug('daemon disabled ; skipped')
        return
    # parameters
    n_workers = getattr(jedi_config.daemon, 'n_proc', 1)
    worker_lifetime = getattr(jedi_config.daemon, 'proc_lifetime', 28800)
    # start
    agent = DaemonMaster(logger=tmp_log,
                            n_workers=n_workers,
                            worker_lifetime=worker_lifetime,
                            tbuf=taskBufferIF, ddmif=ddmIF)
    agent.run()



import traceback

from pandajedi.jedicore.MsgWrapper import MsgWrapper
from pandajedi.jedicore import Interaction
from .TaskSetupperBase import TaskSetupperBase


# logger
from pandacommon.pandalogger.PandaLogger import PandaLogger
logger = PandaLogger().getLogger(__name__.split('.')[-1])


# task setup for general purpose
class SimpleTaskSetupper (TaskSetupperBase):

    # constructor
    def __init__(self,taskBufferIF,ddmIF):
        TaskSetupperBase.__init__(self,taskBufferIF,ddmIF)


    # main to setup task
    def doSetup(self, taskSpec, datasetToRegister, pandaJobs):
        # make logger
        tmpLog = MsgWrapper(logger, "< jediTaskID={0} >".format(taskSpec.jediTaskID))
        tmpLog.info('start label={0} taskType={1}'.format(taskSpec.prodSourceLabel, taskSpec.taskType))
        # returns
        retFatal = self.SC_FATAL
        retOK = self.SC_SUCCEEDED
        try:
            # get DDM I/F
            ddmIF = self.ddmIF.getInterface(taskSpec.vo, taskSpec.cloud)
            # skip if DDM I/F is inactive
            if not ddmIF:
                tmpLog.info('skip due to inactive DDM I/F')
                return retOK
            # collect datasetID to register datasets/containers just in case
            for tmpPandaJob in pandaJobs:
                if not tmpPandaJob.produceUnMerge():
                    for tmpFileSpec in tmpPandaJob.Files:
                        if tmpFileSpec.type in ['output', 'log']:
                            if tmpFileSpec.datasetID not in datasetToRegister:
                                datasetToRegister.append(tmpFileSpec.datasetID)
            # register datasets
            if datasetToRegister:
                tmpLog.info('datasetToRegister={0}'.format(str(datasetToRegister)))
                # get site mapper
                siteMapper = self.taskBufferIF.getSiteMapper()

                # loop over all datasets
                avDatasetList = []
                cnDatasetMap  = {}
                ddmBackEnd = 'rucio'
                for datasetID in datasetToRegister:
                    # get output and log datasets
                    tmpLog.info('getting datasetSpec with datasetID={0}'.format(datasetID))
                    tmpStat, datasetSpec = self.taskBufferIF.getDatasetWithID_JEDI(taskSpec.jediTaskID,
                                                                                   datasetID)
                    if not tmpStat:
                        tmpLog.error('failed to get output and log datasets')
                        return retFatal
                    if datasetSpec.isPseudo():
                        tmpLog.info('skip pseudo dataset')
                        continue

                    tmpLog.info('checking {0}'.format(datasetSpec.datasetName))
                    # check if dataset and container are available in DDM
                    for targetName in [datasetSpec.datasetName, datasetSpec.containerName]:
                        if not targetName:
                            continue
                        if targetName in avDatasetList:
                            tmpLog.info('{0} already registered'.format(targetName))
                            continue
                        # set lifetime
                        lifetime = None
                        # check dataset/container in DDM
                        tmpList = ddmIF.listDatasets(targetName)
                        if not tmpList:
                            # get location
                            location = None
                            locForRule = None
                            if targetName == datasetSpec.datasetName:
                                # dataset
                                tmpLog.info('dest={0}'.format(datasetSpec.destination))
                                if datasetSpec.destination:
                                    if siteMapper.checkSite(datasetSpec.destination):
                                        location = siteMapper.getSite('BNL_OSG_SPHENIX').ddm_output['default']
                                    else:
                                        location = datasetSpec.destination
                            if locForRule is None:
                                locForRule = location
                            # set metadata
                            if targetName == datasetSpec.datasetName:
                                metaData = {}
                                metaData['task_id'] = taskSpec.jediTaskID
                                if taskSpec.campaign:
                                    metaData['campaign'] = taskSpec.campaign
                            else:
                                metaData = None
                            # register dataset/container
                            tmpLog.info('registering {0} with location={1} backend={2} lifetime={3} meta={4}'.format(targetName,
                                                                                                                     location,
                                                                                                                     ddmBackEnd,
                                                                                                                     lifetime,
                                                                                                                     str(metaData)))
                            tmpStat = ddmIF.registerNewDataset(targetName, backEnd=ddmBackEnd, location=location,
                                                               lifetime=lifetime, metaData=metaData)
                            if not tmpStat:
                                tmpLog.error('failed to register {0}'.format(targetName))
                                return retFatal
                            # register location
                            if locForRule:
                                """
                                if taskSpec.workingGroup:
                                    userName = taskSpec.workingGroup
                                else:
                                    userName = taskSpec.userName
                                """
                                userName = None
                                activity = None
                                grouping = None
                                tmpLog.info('registering location={} lifetime={} days activity={} grouping={} '
                                                    'owner={}'.format(locForRule, lifetime, activity, grouping,
                                                                      userName))
                                tmpStat = ddmIF.registerDatasetLocation(targetName, locForRule, owner=userName,
                                                                        lifetime=lifetime, backEnd=ddmBackEnd,
                                                                        activity=activity, grouping=grouping)
                                if not tmpStat:
                                    tmpLog.error('failed to register location {0} for {1}'.format(locForRule,
                                                                                                  targetName))
                                    return retFatal
                            avDatasetList.append(targetName)

                    # check if dataset is in the container
                    if datasetSpec.containerName and datasetSpec.containerName != datasetSpec.datasetName:
                        # get list of constituent datasets in the container
                        if datasetSpec.containerName not in cnDatasetMap:
                            cnDatasetMap[datasetSpec.containerName] = ddmIF.listDatasetsInContainer(datasetSpec.containerName)
                        # add dataset
                        if datasetSpec.datasetName not in cnDatasetMap[datasetSpec.containerName]:
                            tmpLog.info('adding {0} to {1}'.format(datasetSpec.datasetName, datasetSpec.containerName))
                            tmpStat = ddmIF.addDatasetsToContainer(datasetSpec.containerName, [datasetSpec.datasetName],
                                                                   backEnd=ddmBackEnd)
                            if not tmpStat:
                                tmpLog.error('failed to add {0} to {1}'.format(datasetSpec.datasetName,
                                                                               datasetSpec.containerName))
                                return retFatal
                            cnDatasetMap[datasetSpec.containerName].append(datasetSpec.datasetName)
                        else:
                            tmpLog.info('{0} already in {1}'.format(datasetSpec.datasetName, datasetSpec.containerName))
                    # update dataset
                    datasetSpec.status = 'registered'
                    self.taskBufferIF.updateDataset_JEDI(datasetSpec, {'jediTaskID': taskSpec.jediTaskID,
                                                                       'datasetID': datasetID})
            # return
            tmpLog.info('done')
            return retOK
        except Exception as e:
            errStr = 'doSetup failed with {}'.format(str(e))
            tmpLog.error(errStr + traceback.format_exc())
            taskSpec.setErrDiag(errStr)
            return retFatal



# logger
from pandacommon.pandalogger.PandaLogger import PandaLogger
from pandajedi.jedicore.MsgWrapper import MsgWrapper

from pandajedi.jedicore.JediTaskBufferInterface import JediTaskBufferInterface

from pandajedi.jediddm.DDMInterface import DDMInterface

from pandajedi.jediorder.JobBroker import JobBroker
from pandajedi.jediorder.JobSplitter import JobSplitter
from pandajedi.jediorder.JobGenerator import JobGeneratorThread
from pandajedi.jedicore.ThreadUtils import ThreadPool
from pandajedi.jediorder.TaskSetupper import TaskSetupper

import sys

logger = PandaLogger().getLogger('JobGenerator')
tmpLog = MsgWrapper(logger)

tbIF = JediTaskBufferInterface()
tbIF.setupInterface()

siteMapper = tbIF.getSiteMapper()

ddmIF = DDMInterface()
ddmIF.setupInterface()

jediTaskID = int(sys.argv[1])

datasetIDs = None
if len(sys.argv) > 2:
    datasetIDs = [int(sys.argv[2])]

s,taskSpec = tbIF.getTaskWithID_JEDI(jediTaskID)

cloudName = taskSpec.cloud
vo = taskSpec.vo
prodSourceLabel = taskSpec.prodSourceLabel
queueID = taskSpec.workQueue_ID
gshare_name = taskSpec.gshare

workQueue = tbIF.getWorkQueueMap().getQueueWithID(queueID, gshare_name)

threadPool = ThreadPool()

# get typical number of files
#typicalNumFilesMap = tbIF.getTypicalNumInput_JEDI(vo,prodSourceLabel,workQueue,
#                                                  useResultCache=600)

typicalNumFilesMap = {}

tmpListList = tbIF.getTasksToBeProcessed_JEDI(None,vo,workQueue,
                                              prodSourceLabel,
                                              cloudName,nFiles=10,simTasks=[jediTaskID],
                                              fullSimulation=True,
                                              typicalNumFilesMap=typicalNumFilesMap,
                                              simDatasets=datasetIDs)

taskSetupper = TaskSetupper(vo,prodSourceLabel)
taskSetupper.initializeMods(tbIF,ddmIF)

for dummyID,tmpList in tmpListList:
    for taskSpec,cloudName,inputChunk in tmpList:
        jobBroker = JobBroker(taskSpec.vo,taskSpec.prodSourceLabel)
        tmpStat = jobBroker.initializeMods(ddmIF.getInterface(vo),tbIF)
        splitter = JobSplitter()
        gen = JobGeneratorThread(None,threadPool,tbIF,ddmIF,siteMapper,False,taskSetupper,None,None,'dummy',None,None)

        taskParamMap = None
        if taskSpec.useLimitedSites():
            tmpStat,taskParamMap = gen.readTaskParams(taskSpec,taskParamMap,tmpLog)

        tmpStat,inputChunk = jobBroker.doBrokerage(taskSpec,cloudName,inputChunk,taskParamMap)

        #tmpStat,subChunks = splitter.doSplit(taskSpec,inputChunk,siteMapper)



from taskbuffer.Initializer import initializer
from pandajedi.jedicore import JediTaskBuffer

# initialize cx_Oracle using dummy connection
initializer.init()

taskBuffer= JediTaskBuffer.JediTaskBuffer(None)
proxy = taskBuffer.proxyPool.getProxy()
proxy.refreshWorkQueueMap()

#print proxy.workQueueMap.dump()
"""
print proxy.workQueueMap.getQueueWithSelParams('atlas','managed',
                                  prodSourceLabel='managed',
                                  workingGroup='GP_Top')[0].queue_name
print proxy.workQueueMap.getQueueWithSelParams('atlas','managed',
                                  prodSourceLabel='managed',
                                  workingGroup='AP_Top',
                                  processingType='simul')[0].queue_name
print proxy.workQueueMap.getShares('atlas','managed',[1,2,7,8])
"""



#
#
# Setup prog for JEDI
#
#
# set PYTHONPATH to use the current directory first
import sys
sys.path.insert(0,'.')  # noqa: E402

import os
import re
import sys
import socket
import getpass
import grp
import site

import PandaPkgInfo
from setuptools import setup
from setuptools.command.install import install as install_org
import distutils
from distutils.command.install_data import install_data as install_data_org

# define user name and group
panda_user = 'atlpan'
panda_group = 'zp'

# get release version
release_version = PandaPkgInfo.release_version
if 'BUILD_NUMBER' in os.environ:
    release_version = '{0}.{1}'.format(release_version,os.environ['BUILD_NUMBER'])

# get panda specific params
optPanda = {}
newArgv  = []
idx = 0
while idx < len(sys.argv):
    tmpArg = sys.argv[idx]
    if tmpArg.startswith('--panda_'):
        # panda params
        idx += 1
        if len(tmpArg.split('=')) == 2:
            # split to par and val if = is contained
            tmpVal = tmpArg.split('=')[-1]
            tmpArg = tmpArg.split('=')[0]
        elif len(tmpArg.split('=')) == 1:
            tmpVal = sys.argv[idx]
            idx += 1
        else:
            raise RuntimeError("invalid panda option : %s" % tmpArg)
        # get key
        tmpKey = re.sub('--panda_','',tmpArg)
        # set params
        optPanda[tmpKey] = tmpVal
    else:
        # normal opts
        idx += 1
        newArgv.append(tmpArg)
# set new argv
sys.argv = newArgv


# set overall prefix for bdist_rpm
class install_panda(install_org):
    def initialize_options (self):
        install_org.initialize_options(self)

    # disable egg
    def finalize_options(self):
        install_org.finalize_options(self)
        self.single_version_externally_managed = True


# generates files using templates and install them
class install_data_panda (install_data_org):

    def initialize_options (self):
        install_data_org.initialize_options (self)
        self.install_purelib = None
        self.panda_user = panda_user
        self.panda_group = panda_group
        self.virtual_env = ''
        self.virtual_env_setup = ''
        if 'VIRTUAL_ENV' in os.environ:
            self.virtual_env = os.environ['VIRTUAL_ENV']
            self.virtual_env_setup = 'source {0}/bin/activate'.format(os.environ['VIRTUAL_ENV'])
        elif sys.executable:
            venv_dir = os.path.dirname(os.path.dirname(sys.executable))
            py_venv_activate = os.path.join(venv_dir, 'bin/activate')
            if os.path.exists(py_venv_activate):
                self.virtual_env = venv_dir
                self.virtual_env_setup = 'source {0}'.format(py_venv_activate)

    def finalize_options (self):
        # set install_purelib
        self.set_undefined_options('install',
                                   ('install_purelib','install_purelib'))
        # set prefix for pip
        if not hasattr(self, 'prefix'):
            self.prefix = site.PREFIXES[0]
        # set reaming params
        install_data_org.finalize_options(self)
        # set hostname
        if 'hostname' in optPanda and optPanda['hostname'] != '':
            self.hostname = optPanda['hostname']
        else:
            self.hostname = socket.gethostname()
        # set user and group
        if 'username' in optPanda and optPanda['username'] != '':
            self.username = optPanda['username']
        else:
            self.username = getpass.getuser()
        if 'usergroup' in optPanda and optPanda['usergroup'] != '':
            self.usergroup = optPanda['usergroup']
        else:
            self.usergroup = grp.getgrnam(self.username).gr_name

    def run (self):
        # setup.py install sets install_dir to /usr
        if self.install_dir == '/usr':
            self.install_dir = '/'
        elif 'bdist_wheel' in self.distribution.get_cmdline_options():
            # wheel
            self.install_dir = self.prefix
            self.install_purelib = distutils.sysconfig.get_python_lib()
            self.install_scripts = os.path.join(self.prefix, 'bin')
        if not self.install_dir:
            if self.root:
                # rpm
                self.install_dir = self.prefix
                self.install_purelib = distutils.sysconfig.get_python_lib()
                self.install_scripts = os.path.join(self.prefix, 'bin')
            else:
                # sdist
                if not self.prefix:
                    if '--user' in self.distribution.script_args:
                        self.install_dir = site.USER_BASE
                    else:
                        self.install_dir = site.PREFIXES[0]
                else:
                    self.install_dir = self.prefix
        # remove /usr for bdist/bdist_rpm
        match = re.search('(build/[^/]+/dumb)/usr',self.install_dir)
        if match is not None:
            self.install_dir = re.sub(match.group(0),match.group(1),self.install_dir)
        # remove /var/tmp/*-buildroot for bdist_rpm
        match = re.search('(/var/tmp/.*-buildroot)/usr',self.install_dir)
        if match is not None:
            self.install_dir = re.sub(match.group(0),match.group(1),self.install_dir)
        # create tmp area
        tmpDir = 'build/tmp'
        self.mkpath(tmpDir)
        new_data_files = []
        for destDir,dataFiles in self.data_files:
            newFilesList = []
            for srcFile in dataFiles:
                # check extension
                if not srcFile.endswith('.template'):
                    raise RuntimeError("%s doesn't have the .template extension" % srcFile)
                # dest filename
                destFile = re.sub('(\.exe)*\.template$','',srcFile)
                destFile = re.sub(r'^templates/', '', destFile)
                destFile = '%s/%s' % (tmpDir,destFile)
                # open src
                inFile = open(srcFile)
                # read
                filedata=inFile.read()
                # close
                inFile.close()
                # replace patterns
                for item in re.findall('@@([^@]+)@@',filedata):
                    if not hasattr(self,item):
                        raise RuntimeError('unknown pattern %s in %s' % (item,srcFile))
                    # get pattern
                    patt = getattr(self,item)
                    # remove install root, if any
                    if self.root is not None and patt.startswith(self.root):
                        patt = patt[len(self.root):]
                    # remove build/*/dump for bdist
                    patt = re.sub('build/[^/]+/dumb','',patt)
                    # remove /var/tmp/*-buildroot for bdist_rpm
                    patt = re.sub('/var/tmp/.*-buildroot','',patt)
                    # replace
                    filedata = filedata.replace('@@%s@@' % item, patt)
                # write to dest
                if '/' in destFile:
                    destSubDir = os.path.dirname(destFile)
                    if not os.path.exists(destSubDir):
                        os.makedirs(destSubDir)
                oFile = open(destFile,'w')
                oFile.write(filedata)
                oFile.close()
                # chmod for exe
                if srcFile.endswith('.exe.template'):
                    os.chmod(destFile, 0o755)
                # append
                newFilesList.append(destFile)
            # replace dataFiles to install generated file
            new_data_files.append((destDir,newFilesList))
        # install
        self.data_files = new_data_files
        install_data_org.run(self)


# setup for distutils
setup(
    name="panda-jedi",
    version=release_version,
    description='JEDI Package',
    long_description='''This package contains JEDI components''',
    license='GPL',
    author='Panda Team',
    author_email='atlas-adc-panda@cern.ch',
    url='https://twiki.cern.ch/twiki/bin/view/Atlas/PanDA',
    packages=[ 'pandajedi',
               'pandajedi.jedicore',
               'pandajedi.jediexec',
               'pandajedi.jeditest',
               'pandajedi.jedidog',
               'pandajedi.jediddm',
               'pandajedi.jedigen',
               'pandajedi.jedisetup',
               'pandajedi.jediorder',
               'pandajedi.jediconfig',
               'pandajedi.jedirefine',
               'pandajedi.jedithrottle',
               'pandajedi.jedibrokerage',
               'pandajedi.jedipprocess',
               'pandajedi.jedimsgprocessor',
              ],
    install_requires=[
        'six',
        'panda-common',
        'panda-server',
        'python-daemon',
        'numpy',
        'pyyaml',
        'requests',
    ],
    extras_require={
        'oracle': ['cx_Oracle'],
        'mysql': ['mysqlclient'],
        'postgres': ['psycopg2-binary'],
        'rucio': ['rucio-clients'],
        'atlasprod': ['cx_Oracle', 'rucio-clients', 'idds-common', 'idds-client'],
    },
    data_files=[
                # config and cron files
                ('etc/panda', ['templates/panda_jedi.cfg.rpmnew.template',
                                'templates/panda_jedi.cron.template',
                                'templates/logrotate.d/panda_jedi.template',
                               ]
                 ),
                # sysconfig
                ('etc/sysconfig', ['templates/sysconfig/panda_jedi.template',
                                   ]
                 ),
                # init script
                ('etc/rc.d/init.d', ['templates/init.d/panda_jedi.exe.template',
                                   ]
                 ),
                # exec
                ('usr/bin', ['templates/panda_jedi-reniceJEDI.exe.template',
                             ]
                 ),
                ],
    cmdclass={'install': install_panda,
              'install_data': install_data_panda}
)



import sys
from pandajedi.jedicore.JediTaskBufferInterface import JediTaskBufferInterface
from pandajedi.jediddm.DDMInterface import DDMInterface

import multiprocessing

from pandajedi.jediorder import JobGenerator

tbIF = JediTaskBufferInterface()
tbIF.setupInterface()


ddmIF = DDMInterface()
ddmIF.setupInterface()


parent_conn, child_conn = multiprocessing.Pipe()

try:
    testVO = sys.argv[1]
except Exception:
    testVO = 'atlas'

try:
    testTaskType = sys.argv[2]
except Exception:
    testTaskType = 'test'

try:
    execJob = False
    if sys.argv[3] == 'y':
        execJob = True
except Exception:
    pass

try:
    testClouds = sys.argv[4].split(',')
except Exception:
    testClouds = [None]

print('{0} {1} {2}'.format(testVO, testTaskType, testClouds))

gen = multiprocessing.Process(target=JobGenerator.launcher,
                              args=(child_conn,tbIF,ddmIF,testVO,testTaskType,testClouds,
                                    False,execJob))
gen.start()



import sys
import uuid

from pandajedi.jedicore.JediTaskBufferInterface import JediTaskBufferInterface

from pandajedi.jedicore.JediTaskSpec import JediTaskSpec
from pandajedi.jedicore.JediDatasetSpec import JediDatasetSpec

tbIF = JediTaskBufferInterface()
tbIF.setupInterface()

task = JediTaskSpec()
task.jediTaskID = sys.argv[1]
task.taskName = 'pandatest.{0}'.format(uuid.uuid4())
task.status = 'defined'
task.userName = 'pandasrv1'
task.vo = 'atlas'
task.prodSourceLabel = 'managed'
task.taskPriority = 100
task.currentPriority = task.taskPriority
task.architecture = 'i686-slc5-gcc43-opt'
task.transUses = 'Atlas-17.2.7'
task.transHome = 'AtlasProduction-17.2.8.10'
task.transPath = 'Reco_trf.py'
task.workQueue_ID = 3

tbIF.insertTask_JEDI(task)


ds = JediDatasetSpec()
ds.jediTaskID = task.jediTaskID
if len(sys.argv) > 2:
    ds.datasetName = sys.argv[2]
else:
    ds.datasetName = 'data12_8TeV.00214651.physics_Egamma.merge.AOD.f489_m1261'
ds.type = 'input'
ds.vo = task.vo
ds.cloud = 'US'
ds.streamName = 'IN'
ds.status = 'defined'
ds.nFiles = 0
ds.nFilesUsed = 0
ds.nFilesFinished = 0
ds.nFilesFailed = 0

st,datasetID = tbIF.insertDataset_JEDI(ds)

ds = JediDatasetSpec()
ds.jediTaskID = task.jediTaskID
ds.datasetName = 'ddo.000001.Atlas.Ideal.DBRelease.v220701'
ds.type = 'input'
ds.vo = task.vo
ds.cloud = 'US'
ds.streamName = 'DBR'
ds.status = 'defined'
ds.nFiles = 0
ds.nFilesUsed = 0
ds.nFilesFinished = 0
ds.nFilesFailed = 0
ds.masterID = datasetID
ds.attributes = 'repeat,nosplit'

tbIF.insertDataset_JEDI(ds)


ds = JediDatasetSpec()
ds.jediTaskID = task.jediTaskID
ds.datasetName = 'panda.jeditest.GEN.{0}'.format(uuid.uuid4())
ds.type = 'output'
ds.vo = task.vo
ds.cloud = 'US'
ds.streamName = 'OUT'
ds.status = 'defined'
ds.nFiles = 0
ds.nFilesUsed = 0
ds.nFilesFinished = 0
ds.nFilesFailed = 0

st,datasetID = tbIF.insertDataset_JEDI(ds)

tbIF.insertOutputTemplate_JEDI([{'jediTaskID':task.jediTaskID,
                                 'datasetID':datasetID,
                                 'filenameTemplate':'{0}.${{SN}}.pool.root'.format(ds.datasetName),
                                 'serialNr':1,
                                 'streamName':'OUT',
                                 'outtype':ds.type}])

ds = JediDatasetSpec()
ds.jediTaskID = task.jediTaskID
ds.datasetName = 'panda.jeditest.log.{0}'.format(uuid.uuid4())
ds.type = 'log'
ds.vo = task.vo
ds.cloud = 'US'
ds.streamName = 'LOG'
ds.status = 'defined'
ds.nFiles = 0
ds.nFilesUsed = 0
ds.nFilesFinished = 0
ds.nFilesFailed = 0

st,datasetID = tbIF.insertDataset_JEDI(ds)

tbIF.insertOutputTemplate_JEDI([{'jediTaskID':task.jediTaskID,
                                 'datasetID':datasetID,
                                 'filenameTemplate':'{0}.${{SN}}.log.tgz'.format(ds.datasetName),
                                 'serialNr':1,
                                 'streamName':'LOG',
                                 'outtype':ds.type}])

tbIF.insertJobParamsTemplate_JEDI(task.jediTaskID,'inputAODFile=${IN} maxEvents=1000 RunNumber=213816 autoConfiguration=everything preExec="from BTagging.BTaggingFlags import BTaggingFlags;BTaggingFlags.CalibrationTag=\"BTagCalibALL-07-02\"" DBRelease=${DBR} AMITag=p1462 outputNTUP_EMBLLDNFile=${OUT}')



"""
dataset specification for JEDI

"""

import re
import math

from pandajedi.jediconfig import jedi_config


class JediDatasetSpec(object):
    
    def __str__(self):
        sb = []
        for key in self.__dict__:
            if key == 'Files':
                sb.append("{key}='{value}'".format(key=key, value=len(self.__dict__[key])))
            else:
                sb.append("{key}='{value}'".format(key=key, value=self.__dict__[key]))
        return ', '.join(sb)
    
    def __repr__(self):
        return self.__str__() 
    
    # attributes
    _attributes = (
        'jediTaskID','datasetID','datasetName','containerName',
        'type','creationTime','modificationTime','vo','cloud',
        'site','masterID','provenanceID','status','state',
        'stateCheckTime','stateCheckExpiration','frozenTime',
        'nFiles','nFilesToBeUsed','nFilesUsed',
        'nFilesFinished','nFilesFailed','nFilesOnHold',
        'nEvents','nEventsToBeUsed','nEventsUsed',
        'lockedBy','lockedTime','attributes','streamName',
        'storageToken','destination','templateID','nFilesWaiting'
        )
    # attributes which have 0 by default
    _zeroAttrs = ()
    # attributes to force update
    _forceUpdateAttrs = ('lockedBy','lockedTime')
    # mapping between sequence and attr
    _seqAttrMap = {'datasetID':'{0}.JEDI_DATASETS_ID_SEQ.nextval'.format(jedi_config.db.schemaJEDI)}
    # token for attributes
    attrToken = {
        'allowNoOutput'    : 'an',
        'consistencyCheck' : 'cc',
        'eventRatio'       : 'er',
        'indexConsistent'  : 'ic',
        'mergeOnly'        : 'mo',
        'nFilesPerJob'     : 'np',
        'num_records'      : 'nr',
        'offset'           : 'of',
        'objectStore'      : 'os',
        'pseudo'           : 'ps',
        'random'           : 'rd',
        'reusable'         : 'ru',
        'transient'        : 'tr',
        'useDuplicated'    : 'ud',
        }


    # constructor
    def __init__(self):
        # install attributes
        for attr in self._attributes:
            object.__setattr__(self,attr,None)
        # file list
        object.__setattr__(self,'Files',[])
        # map of changed attributes
        object.__setattr__(self,'_changedAttrs',{})
        # distributed
        object.__setattr__(self,'distributed',False)



    # override __setattr__ to collecte the changed attributes
    def __setattr__(self,name,value):
        oldVal = getattr(self,name)
        object.__setattr__(self,name,value)
        newVal = getattr(self,name)
        # collect changed attributes
        if oldVal != newVal or name in self._forceUpdateAttrs:
            self._changedAttrs[name] = value



    # add File to files list
    def addFile(self,fileSpec):
        # append
        self.Files.append(fileSpec)


        
    # reset changed attribute list
    def resetChangedList(self):
        object.__setattr__(self,'_changedAttrs',{})



    # force update
    def forceUpdate(self,name):
        if name in self._attributes:
            self._changedAttrs[name] = getattr(self,name)
        

    
    # return map of values
    def valuesMap(self,useSeq=False,onlyChanged=False):
        ret = {}
        for attr in self._attributes:
            # use sequence
            if useSeq and attr in self._seqAttrMap:
                continue
            # only changed attributes
            if onlyChanged:
                if attr not in self._changedAttrs:
                    continue
            val = getattr(self,attr)
            if val is None:
                if attr in self._zeroAttrs:
                    val = 0
                else:
                    val = None
            ret[':%s' % attr] = val
        return ret



    # pack tuple into FileSpec
    def pack(self,values):
        for i in range(len(self._attributes)):
            attr= self._attributes[i]
            val = values[i]
            object.__setattr__(self,attr,val)



    # return column names for INSERT
    def columnNames(cls,prefix=None):
        ret = ""
        for attr in cls._attributes:
            if prefix is not None:
                ret += '{0}.'.format(prefix)
            ret += '{0},'.format(attr)
        ret = ret[:-1]    
        return ret
    columnNames = classmethod(columnNames)



    # return expression of bind variables for INSERT
    def bindValuesExpression(cls,useSeq=True):
        ret = "VALUES("
        for attr in cls._attributes:
            if useSeq and attr in cls._seqAttrMap:
                ret += "%s," % cls._seqAttrMap[attr]
            else:
                ret += ":%s," % attr
        ret = ret[:-1]
        ret += ")"            
        return ret
    bindValuesExpression = classmethod(bindValuesExpression)


    
    # return an expression of bind variables for UPDATE to update only changed attributes
    def bindUpdateChangesExpression(self):
        ret = ""
        for attr in self._attributes:
            if attr in self._changedAttrs:
                ret += '%s=:%s,' % (attr,attr)
        ret  = ret[:-1]
        ret += ' '
        return ret



    # set dataset attribute
    def setDatasetAttribute(self,attr):
        if self.attributes is None:
            self.attributes = ''
        else:
            self.attributes += ','
        self.attributes += attr



    # set dataset attribute with label
    def setDatasetAttributeWithLabel(self,label):
        if label not in self.attrToken:
            return
        attr = self.attrToken[label]
        if self.attributes is None:
            self.attributes = ''
        else:
            self.attributes += ','
        self.attributes += attr

    

    # get the total size of files
    def getSize(self):
        totalSize = 0
        checkedList = []
        for tmpFileSpec in self.Files:
            if tmpFileSpec.lfn not in checkedList:
                totalSize += tmpFileSpec.fsize
                checkedList.append(tmpFileSpec.lfn)
        return totalSize    



    # return list of status to update contents
    def statusToUpdateContents(cls):
        return ['defined','toupdate']
    statusToUpdateContents = classmethod(statusToUpdateContents)



    # return list of types for input
    def getInputTypes(cls):
        return ['input','pseudo_input']
    getInputTypes = classmethod(getInputTypes)



    # return list of types to generate jobs
    def getProcessTypes(cls):
        return cls.getInputTypes() + ['pp_input'] + cls.getMergeProcessTypes()
    getProcessTypes = classmethod(getProcessTypes)



    # return list of types for merging
    def getMergeProcessTypes(cls):
        return ['trn_log','trn_output']
    getMergeProcessTypes = classmethod(getMergeProcessTypes)


    
    # get type of unkown input
    def getUnknownInputType(cls):
        return 'trn_unknown'
    getUnknownInputType = classmethod(getUnknownInputType)



    # check if JEDI needs to keep track of file usage
    def toKeepTrack(self):
        if self.isNoSplit() and self.isRepeated():
            return False
        elif self.isReusable():
            return False
        else:
            return True



    # check if it is not split
    def isNoSplit(self):
        if self.attributes is not None and 'nosplit' in self.attributes:
            return True
        else:
            return False



    # check if it is repeatedly used
    def isRepeated(self):
        if self.attributes is not None and 'repeat' in self.attributes:
            return True
        else:
            return False



    # check if it is randomly used
    def isRandom(self):
        if self.attributes is not None and 'rd' in self.attributes.split(','):
            return True
        else:
            return False



    # check if it is reusable
    def isReusable(self):
        if self.attributes is not None and 'ru' in self.attributes.split(','):
            return True
        else:
            return False



    # check if consistency is checked
    def checkConsistency(self):
        if self.attributes is not None and 'cc' in self.attributes.split(','):
            return True
        else:
            return False



    # set consistency is checked
    def enableCheckConsistency(self):
        if self.attributes in [None,'']:
            self.attributes = 'cc'
        elif 'cc' not in self.attributes.split(','):
            self.attributes += ',cc'



    # check if it is pseudo
    def isPseudo(self):
        if self.datasetName in ['pseudo_dataset','seq_number'] \
                or self.type in ['pp_input']:
            return True
        if self.attributes is not None and self.attrToken['pseudo'] in self.attributes.split(','):
            return True
        return False



    # check if it is a many-time dataset which is treated as long-standing at T2s
    def isManyTime(self):
        if self.attributes is not None and 'manytime' in self.attributes:
            return True
        else:
            return False



    # check if it is seq number
    def isSeqNumber(self):
        if self.datasetName in ['seq_number']:
            return True
        else:
            return False



    # check if duplicated files are used
    def useDuplicatedFiles(self):
        if self.attributes is not None and ('usedup' in self.attributes or \
                                            'ud' in  self.attributes.split(',')):
            return True
        else:
            return False



    # check if it is a master dataset
    def isMaster(self):
        if self.masterID is None and self.type in self.getProcessTypes():
            return True
        else:
            return False



    # check if it is a master input dataset
    def isMasterInput(self):
        if self.masterID is None and self.type in self.getInputTypes():
            return True
        else:
            return False



    # remove nosplit attribute
    def remAttribute(self,attrName):
        if self.attributes is not None:
            self.attributes = re.sub(attrName,'',self.attributes)
            self.attributes = re.sub(',,',',',self.attributes)
            self.attributes = re.sub('^,','',self.attributes)
            self.attributes = re.sub(',$','',self.attributes)
            if self.attributes == '':
                self.attributes = None
        


    # remove nosplit attribute
    def remNoSplit(self):
        self.remAttribute('nosplit')



    # remove repeat attribute
    def remRepeat(self):
        self.remAttribute('repeat')



    # get the ratio to master
    def getRatioToMaster(self):
        try:
            tmpMatch = re.search('ratio=(\d+(\.\d+)*)',self.attributes)
            if tmpMatch is not None:
                ratioStr = tmpMatch.group(1)
                try:
                    # integer
                    return int(ratioStr)
                except Exception:
                    pass
                try:
                    # float
                    return float(ratioStr)
                except Exception:
                    pass
        except Exception:
            pass
        return 1



    # get N multiplied by ratio
    def getNumMultByRatio(self,num):
        # no split
        if self.isNoSplit():
            return None
        # get ratio
        ratioVal = self.getRatioToMaster()
        # integer or float
        if isinstance(ratioVal,int):
            retVal = num * ratioVal
        else:
            retVal = float(num) * ratioVal
            retVal = int(math.ceil(retVal))
        return retVal



    # unique map key for output
    def outputMapKey(self):
        mapKey = '{0}#{1}'.format(self.datasetName,self.provenanceID)
        return mapKey



    # unique map key
    def uniqueMapKey(self):
        mapKey = '{0}#{1}'.format(self.datasetName,self.datasetID)
        return mapKey



    # set offset
    def setOffset(self,offset):
        self.setDatasetAttribute('{0}={1}'.format(self.attrToken['offset'],offset))



    # get offset
    def getOffset(self):
        if self.attributes is not None:
            tmpMatch = re.search(self.attrToken['offset']+'=(\d+)',self.attributes)
            if tmpMatch is not None:
                offset = int(tmpMatch.group(1))
                return offset
        return 0



    # set number of records
    def setNumRecords(self,n):
        self.setDatasetAttribute('{0}={1}'.format(self.attrToken['num_records'],n))



    # get number of records
    def getNumRecords(self):
        if self.attributes is not None:
            for item in self.attributes.split(','):
                tmpMatch = re.search(self.attrToken['num_records']+'=(\d+)',item)
                if tmpMatch is not None:
                    num_records = int(tmpMatch.group(1))
                    return num_records
        return None



    # set object store
    def setObjectStore(self,objectStore):
        self.setDatasetAttribute('{0}={1}'.format(self.attrToken['objectStore'],objectStore))



    # get object store
    def getObjectStore(self):
        if self.attributes is not None:
            tmpMatch = re.search(self.attrToken['objectStore']+'=([^,]+)',self.attributes)
            if tmpMatch is not None:
                return tmpMatch.group(1)
        return None



     # set the number of files per job
    def setNumFilesPerJob(self,num):
        self.setDatasetAttribute('{0}={1}'.format(self.attrToken['nFilesPerJob'],num))



    # get the number of files per job
    def getNumFilesPerJob(self):
        if self.attributes is not None:
            tmpMatch = re.search(self.attrToken['nFilesPerJob']+'=(\d+)',self.attributes)
            if tmpMatch is not None:
                num = int(tmpMatch.group(1))
                return num
        # use continuous numbers for seq_number
        if self.isSeqNumber():
            return 1
        return None



    # check if unmerged dataset
    def toMerge(self):
        if self.type.startswith('trn_'):
            return True
        return False



    # set transient
    def setTransient(self,val):
        if val is True:
            val = 1
        else:
            val = 0
        self.setDatasetAttribute('{0}={1}'.format(self.attrToken['transient'],val))



    # get transient
    def getTransient(self):
        if self.attributes is not None:
            for item in self.attributes.split(','):
                tmpMatch = re.search(self.attrToken['transient']+'=(\d+)',item)
                if tmpMatch is not None:
                    val = int(tmpMatch.group(1))
                    if val == 1:
                        return True
                    else:
                        return False
        return None



    # check if no output is allowed
    def isAllowedNoOutput(self):
        if self.attributes is not None and self.attrToken['allowNoOutput'] in self.attributes.split(','):
            return True
        else:
            return False



    # allow no output
    def allowNoOutput(self):
        if self.attributes in [None,'']:
            items = []
        else:
            items = self.attributes.split(',')
        if self.attrToken['allowNoOutput'] not in items:
            items.append(self.attrToken['allowNoOutput'])
            self.attributes = ','.join(items)



    # check if index consistency is required
    def indexConsistent(self):
        if self.attributes is not None and self.attrToken['indexConsistent'] in self.attributes.split(','):
            return True
        else:
            return False



    # set distributed
    def setDistributed(self):
        self.distributed = True

    # reset distributed
    def reset_distributed(self):
        self.distributed = False

    # check if distributed
    def isDistributed(self):
        return self.distributed



    # set event ratio
    def setEventRatio(self,num):
        self.setDatasetAttribute('{0}={1}'.format(self.attrToken['eventRatio'],num))



    # get event ratio
    def getEventRatio(self):
        if self.attributes is not None:
            for item in self.attributes.split(','):
                tmpMatch = re.search(self.attrToken['eventRatio']+'=(\d+(\.\d+)*)',item)
                if tmpMatch is not None:
                    ratioStr = tmpMatch.group(1)
                    try:
                        # integer
                        return int(ratioStr)
                    except Exception:
                        pass
                    try:
                        # float
                        return float(ratioStr)
                    except Exception:
                        pass
        return None

    # set pseudo
    def setPseudo(self):
        if self.attributes in [None,'']:
            items = []
        else:
            items = self.attributes.split(',')
        if self.attrToken['pseudo'] not in items:
            items.append(self.attrToken['pseudo'])
            self.attributes = ','.join(items)

    # merge only
    def is_merge_only(self):
        try:
            return self.attrToken['mergeOnly'] in self.attributes.split(',')
        except Exception:
            return False
